@@ -4,127 +4,6 @@
 # This is Open Source software, released under the BSD 2-clause license,
 # see http://opensource.org/licenses/BSD-2-Clause for details.
 
-"""Interface to a yorick process.
-
-yo = Yorick()      start a yorick process
-yo.kill()          kill a yorick process
-
-yo('code')         execute code in yorick, return None
-v = yo('=expr')    execute expr in yorick, return value
-
-Three different handles to the yorick process provide a nicer interface:
-  chandle = yo.call      call-semantics handle
-  ehandle = yo.evaluate  eval-semantics handle
-  vhandle = yo.value     value-semantics handle
-  chandle, ehandle, vhandle = yo.handles(7)
-For interactive use, you may abbreviate
-yo.call as yo.c, yo.evaluate as yo.e, and yo.value as yo.v.
-
-Attributes of any handle object represent yorick variables, for example:
-  yo.v.var = <expr>   sets the yorick variable var to the python <expr>
-  yo.v.var            in a python expression gets the yorick variable var
-The three types of handles come into play when the yorick variable is a
-function rather than data, or when you want to refer to data stored in a
-yorick variable without moving the entire array from yorick to python.
-
-Python has only one syntax for invoking functions, whereas yorick has two
--- one to invoke the function and return its value, the other to invoke
-the function as a subroutine, discarding its value.  The eval-semantics
-handle produces yorick function references which return a value, whereas
-the call-semantics handle produces yorick functions which will be invoked
-as subroutines:
-  yo.e.atan(<expr>)   returns yorick atan(<expr>)
-  yo.c.plg(y, x)      invokes yorick plg,y,x
-You can pass both positional and keyword arguments to either type of
-function reference:
-  yo.c.plg(y, x, color='red')
-
-Thus, you usually want to use yo.v to set or get array data in yorick
-variables, yo.c to call yorick functions as subroutines (discarding any
-result), yo.e to return a yorick function value, and yo('code') to
-parse yorick code (for example, to define an interpreted function).
-The exception is, when you want to set or get only a part of a yorick
-array, because the whole array is very large and you don't want the
-performance penalty of transmitting the whole thing to or from yorick.
-To do that, use the evaluate instead of the value handle:
-  yo.e.var[ndx1, ndx2, ...] = <expr>  # set a slice of yorick array var
-  yo.e.var[ndx1, ndx2, ...]           # get a slice of yorick array var
-Each ndxI expression can be a scalar value, a list of integers, or a
-slice start:stop:step.  The index expressions have yorick semantics, not
-numpy semantics, that is: (A) dimension order is fastest varying to
-slowest varying, (B) index origin is 1, and (C) the stop value in a
-slice is included as part of the slice.  If you want numpy index semantics,
-you can use the call handle yo.c.var[ndxlist], and pyorick will swap the
-index order and attempt to fix the index origin and slice stop values.
-
-A few potential yorick variable names cannot be accessed using the
-yo.<handle>.varname syntax (e.g.- __init__).  For these cases, all
-three handles accept a dict-like key:
-  yo.v['var']   same as   yo.v.var   (same for yo.c, yo.e handles)
-This feature is useful if the name of the yorick variable is the value
-of a python string variable as well.
-
-The call and evaluate handle attributes return a reference to a yorick
-variable, which doesn't actually communicate with yorick until you use
-it.  That is, yo.e.var is just a reference to the yorick variable 'var'
-with eval-sematics, while yo.c.var is a reference with call-semantics.
-Variable reference objects implement methods to actually retrieve data
-with yo.e.var(args) or yo.e.var[ndxs].  Note that yorick does not distinguish
-between var(args) and var[ndxs], so they do the same thing in pyorick.
-However, in python, yo.e.var(1:2) is a syntax error, while yo.e.var[1:2]
-is not.  Similarly, python syntax does not permit keywords in index lists,
-nor keywords preceding positional arguments in argument lists.
-
-Yorick variable reference objects have several properties:
-  yo.e.var.info     # returns datatype and shape information about var
-  yo.e.var.value    # returns the value of var, like yo.v.var
-  yo.e.var.v        # returns the value of var, like yo.v.var
-In general, you can convert any variable handle sematics to another
-semantics, so yor.e.var.c is a call-semantics reference for var,
-yo.c.var.e is and eval-semantics references, and so on.  Info returns
-a 1D array [typeid, rank, dim1, dim2, ..., dimN] for an array type,
-where typeid is 1, 2, 3 for short, int, long integers, 8 for bytes
-(char in yorick, uint8 in numpy), 5, 6 for float, double reals, 14
-for complex, and 16 for string data.  The dimension lengths are in
-yorick order, fastest to slowest varying in memory.  For non-array
-data, info returns a single element array [typeid], -1 for a function,
--2 for a list-like anonymous object, -3 for a dict-like object, -4 for
-a slice, -5 for nil, -6 for a file handle, and -7 or -8 for
-other non-representable objects.
-
-An value handle attribute representing a non-data yorick variable, such
-as a function or a file handle, also returns a yorick variable reference
-(after a brief exchange with yorick).  A reference returned by a value
-handle in this way is treated like a call-semantics reference.  Thus,
-yo.v.plg is essentially the same as yo.c.plg.
-
-All yorick array types except pointer and struct instance are valid data.
-A yorick string maps to a python str, but only str which do not contain
-any '\0' characters are possible in yorick.  Going from python to yorick,
-a str is silently truncated at its first '\0'.  A yorick array of strings
-becomes a nested list of python str.  In general, nested lists of python
-numbers will be converted to numpy arrays and sent to yorick.  A numpy
-array of strings, in addition to a nested list of strings, can also be
-sent to yorick.  A python list (or other sequence) of arbitrary
-representable data objects maps to an anonymous oxy object in yorick.
-A python dict with str keys maps to a yorick oxy object with named
-members.
-
-The following special objects are available for use in expressions used
-to set variable values in yorick:
-  ystring0     yorick string(0) is C NULL, different than ""
-  ynewaxis     np.newaxis is None, which yorick interprets as :
-The ystring0 value is also passed back to python to represent a string(0)
-value in yorick.  It is derived from str and has value '' in python.  You
-can check for it with "s is ystring0" if you need to distinguish.
-
-Finally, pyorick can turn the python command line into a yorick terminal:
-  yo()          enter yorick terminal mode, special yorick commands are:
-    py            from yorick terminal mode returns to python
-    py, "code"    from yorick terminal mode to exec code in python
-    py("expr")    from yorick terminal mode to eval expr in python
-"""
-
 # Attempt to make it work for both python 2.6+ and python 3.x.
 # Avoid both six and future modules, which are often not installed.
 #from __future__ import (absolute_import, division,
@@ -175,23 +54,22 @@ class Yorick(object):
       extra (str or list of str):  additional command line arguments
     """
     if isinstance(extra, Process):
-      self.proc = extra
+      self.bare = YorickBare(extra)
+    elif isinstance(extra, YorickBare):
+      self.bare = extra
     else:
-      self.proc = ProcessDefault(extra, **kwargs)
+      self.bare = YorickBare(ProcessDefault(extra, **kwargs))
     self._call = self._eval = self._value = None
 
-  def __del__(self):
-    self.kill()
-
   def __repr__(self):
-    return "<connection to {0}>".format(repr(self.proc)[1:-1])
+    return "<connection to {0}>".format(repr(self.bare)[1:-1])
 
   def kill(self):
     """Kill yorick process."""
-    self.proc.kill()
+    self.bare.proc.kill()
   def debug(self, on):
     """Set or unset debug mode for yorick process."""
-    self.proc.debug(on)
+    self.bare.proc.debug(on)
 
   def handles(self, which=3):
     """Return handles whose attributes are yorick variables.
@@ -213,27 +91,42 @@ class Yorick(object):
   @property
   def call(self):
     if not self._call:
-      self._call = YorickHandle(0, self)
+      self._call = YorickHandle(0, self.bare)
     return self._call
   @property
   def evaluate(self):
     if not self._eval:
-      self._eval = YorickHandle(1, self)
+      self._eval = YorickHandle(1, self.bare)
     return self._eval
   @property
   def value(self):
     if not self._value:
-      self._value = YorickHandle(2, self)
+      self._value = YorickHandle(2, self.bare)
     return self._value
   # single character abbreviations for interactive use
   c = call
   e = evaluate
   v = value
+  # FIXME: These create circular reference since they own use of self!
+  # fix using weakref module?  not obvious how
+  # - should both Yorick object and YorickHandle obejct refer to a common
+  # underlying object?  YorickVar also retains reference to Yorick
 
   def __call__(self, command=None, *args, **kwargs):  # pipe(command)
     """Execute a yorick command."""
     return self.call(command, *args, **kwargs)
 
+class YorickBare(object):
+  """Avoids circular references among Yorick, YorickHandle, and YorickVar."""
+  def __init__(self, proc):
+    self.proc = proc
+
+  def __del__(self):
+    self.proc.kill()
+
+  def __repr__(self):
+    return "<bare connection to {0}>".format(repr(self.proc)[1:-1])
+
   def _reqrep(self, msgid, *args, **kwargs):  # convenience for YorickHandle
     reply = Message()
     self.proc.reqrep(Message(msgid, *args, **kwargs), reply)
@@ -275,49 +168,49 @@ class YorickHandle(object):
   # However, no other attributes are permitted.
 
   # must avoid __setattr__, __getattr__ by explicit calls through __dict__
-  def __init__(self, reftype, connection):
+  def __init__(self, reftype, bare):
     self.__dict__['_reftype__'] = reftype
-    self.__dict__['_yorick__'] = connection
+    self.__dict__['_yorick__'] = bare
 
   def __repr__(self):
-    connection = self.__dict__['_yorick__']
+    bare = self.__dict__['_yorick__']
     typ = ['call', 'evaluate', 'value'][self.__dict__['_reftype__']]
     s = "<yorick {0}-semantics handle to {1}>"
-    return s.format(typ, repr(connection.proc)[1:-1])
+    return s.format(typ, repr(bare.proc)[1:-1])
 
   def __getitem__(self, key=None):
     """Return parent connection, avoiding use of an attribute."""
-    connection = self.__dict__['_yorick__']
+    bare = self.__dict__['_yorick__']
     if not key:
-      return connection
+      return Yorick(bare)
     typ = self.__dict__['_reftype__']
     if typ == 2:
-      return connection._reqrep(ID_GETVAR, key)
-    return YorickVarDerived(connection, typ, key)
+      return bare._reqrep(ID_GETVAR, key)
+    return YorickVarDerived(bare, typ, key)
 
   def __call__(self, command=None, *args, **kwargs):
     """Implement handle(command) or handle(format, args, key=kwds)."""
     if args or kwargs:
       command = command.format(*args, **kwargs)
-    connection = self.__dict__['_yorick__']
+    bare = self.__dict__['_yorick__']
     typ = self.__dict__['_reftype__']
     if command and command[0]=='=':   # leading = forces eval semantics
       command = command[1:]
       typ = 1
     if command is None:
       print("--> yorick prompt (type py to return to python):")
-      connection.proc.interact(YorickServer())
+      bare.proc.interact(YorickServer())
       print("--> python prompt:")
     elif typ:
-      return connection._reqrep(ID_EVAL, command)
+      return bare._reqrep(ID_EVAL, command)
     else:
-      return connection._reqrep(ID_EXEC, command)
+      return bare._reqrep(ID_EXEC, command)
 
   def __setattr__(self, name, value):
     """Implement handle.name = value."""
-    connection = self.__dict__['_yorick__']
+    bare = self.__dict__['_yorick__']
     if name not in self.__dict__:
-      connection._reqrep(ID_SETVAR, name, value)
+      bare._reqrep(ID_SETVAR, name, value)
     else:
       object.__setattr__(self, name, value)
 
@@ -328,43 +221,43 @@ class YorickHandle(object):
       raise AttributeError("YorickHandle instance has no attribute '"+name+"'")
     if name in ['_reftype__', '_yorick__']:
       return self.__dict__[name]
-    connection = self.__dict__['_yorick__']
+    bare = self.__dict__['_yorick__']
     typ = self.__dict__['_reftype__']
     if typ == 2:
-      return connection._reqrep(ID_GETVAR, name)
-    return YorickVarDerived(connection, typ, name)
+      return bare._reqrep(ID_GETVAR, name)
+    return YorickVarDerived(bare, typ, name)
 
 class YorickVar(object):
   """Reference to a yorick variable."""
-  def __init__(self, yorick, reftype, name):
+  def __init__(self, bare, reftype, name):
     if not isinstance(name, basestring):
       raise PYorickError("illegal yorick variable name")
-    self.yorick = yorick
+    self.bare = bare
     self.reftype = bool(reftype)
     self.name = name
 
   def __repr__(self):
-    yorick = self.yorick
+    bare = self.bare
     typ = ['call', 'evaluate'][self.reftype]
     s = "<yorick variable {0} ({1}) in {2}>"
-    return s.format(self.name, typ, repr(yorick.proc)[1:-1])
+    return s.format(self.name, typ, repr(bare.proc)[1:-1])
 
   def __call__(self, *args, **kwargs):
     """Implement handle.name(args, kwargs)."""
     if self.reftype:
-      return self.yorick._reqrep(ID_FUNCALL, self.name, *args, **kwargs)
+      return self.bare._reqrep(ID_FUNCALL, self.name, *args, **kwargs)
     else:
-      self.yorick._reqrep(ID_SUBCALL, self.name, *args, **kwargs)
+      self.bare._reqrep(ID_SUBCALL, self.name, *args, **kwargs)
 
   def __getitem__(self, key): 
     """Implement handle.name[key]."""
     key = self._fix_indexing(key)
-    return self.yorick._reqrep(ID_GETSLICE, self.name, *key)
+    return self.bare._reqrep(ID_GETSLICE, self.name, *key)
 
   def __setitem__(self, key, value):
     """Implement handle.name[key] = value."""
     key = self._fix_indexing(key) + (value,)
-    return self.yorick._reqrep(ID_SETSLICE, self.name, *key)
+    return self.bare._reqrep(ID_SETSLICE, self.name, *key)
 
   def _fix_indexing(self, key):
     if not isinstance(key, tuple):
@@ -400,12 +293,12 @@ class YorickVar(object):
   @property
   def info(self):
     """Implement handle.name.info."""
-    return self.yorick._reqrep(ID_GETSHAPE, self.name)
+    return self.bare._reqrep(ID_GETSHAPE, self.name)
 
   @property
   def value(self):
     """Implement handle.name.value."""
-    return self.yorick._reqrep(ID_GETVAR, self.name)
+    return self.bare._reqrep(ID_GETVAR, self.name)
   # single character alias for interactive use
   v = value
 
@@ -413,7 +306,7 @@ class YorickVar(object):
   def call(self):
     """Implement handle.name.call."""
     if self.reftype:
-      return YorickVarDerived(self.yorick, False, self.name)
+      return YorickVarDerived(self.bare, False, self.name)
     else:
       return self
   # single character alias for interactive use
@@ -423,7 +316,7 @@ class YorickVar(object):
   def evaluate(self):
     """Implement handle.name.evaluate."""
     if not self.reftype:
-      return YorickVarDerived(self.yorick, True, self.name)
+      return YorickVarDerived(self.bare, True, self.name)
     else:
       return self
   # single character alias for interactive use
@@ -1272,16 +1165,48 @@ def find_package_data(name):
   # to the newline character.
   # This convention makes it straightforward to install pyorick "by hand"
   # when distutils cannot be used.
+  # When pyorick.i0 is not found, check for it in the yorick user directory
+  # ~/.yorick/ (or similar).  If not found there, but pkgutil.get_data finds
+  # it, unpack pyorick.i0 to ~/.yorick/ (or an existing yorick customization
+  # directory).
+  path = None
   try:
-    path = __file__
-    if os.path.islink(path):
-      path = os.path.realpath(path)
-    path = os.path.join(os.path.dirname(os.path.abspath(path)), name)
+    p = __file__
+    if os.path.islink(p):
+      p = os.path.realpath(p)
+    p = os.path.join(os.path.dirname(os.path.abspath(p)), name)
+    if os.path.isfile(p):
+      path = p
+    else:
+      # before giving up, try ~/yorick directories
+      home = os.path.expandpath('~')
+      d = yuser = '.yorick'  # first choice is ~/.yorick
+      for d in [yuser, 'yorick', 'Library/Yorick', 'Application Data/Yorick',
+                'Yorick']:  # possibilities, in order, checked in yorick/std0.c
+        p = os.path.join(home, d)
+        if os.path.isdir(p):
+          yuser = p
+          break
+      p = os.path.join(yuser, name)
+      if not os.path.isfile(p):
+        path = p
+      else:
+        # last ditch effort, needed if pyorick loaded from a zip file
+        import pkgutil
+        d = pkgutil.get_data('pyorick', 'pyorick.i0')
+        if d:
+          d = d.decode('utf-8').splitlines() # removes universal newlines
+          if not os.path.isdir(yuser):
+            os.mkdir(yuser)
+          with open(p, 'w') as f:
+            for line in d:
+              f.write("{0}\n".format(line))
+          path = p
   except:
-    path = 'I am-not-a file, am.I'
-  if not os.path.exists(path):
+    pass
+  if path is None:
     raise PYorickError('unable to find '+name)
-  return path
+  return os.path.normcase(path)
 
 ypathd = "yorick"   # default yorick command
 ipathd = find_package_data("pyorick.i0")  # default pyorick.i0 include file
